// Header
#include "EntitySnapshot.h"

// STD / STL
#include <iostream>
#include <fstream>

// ArmarX
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <RobotAPI/libraries/aron/core/data/variant/container/Dict.h>

// ArmarX
#include "../base/extractor/noExtractor/NoExtractor.h"
#include "../base/extractor/imageExtractor/ImageExtractor.h"
#include "../base/converter/dict/json/JsonConverter.h"
#include "../base/converter/dict/bson/BsonConverter.h"
#include "../base/converter/image/png/PngConverter.h"

namespace armarx::armem::server::ltm::disk
{

    namespace
    {
        MemoryID getMemoryIDFromPath(const std::filesystem::path& p)
        {
            util::ensureFolderExists(p);

            MemoryID m;
            m.memoryName = p.parent_path().parent_path().parent_path().parent_path().filename();
            m.coreSegmentName = p.parent_path().parent_path().parent_path().filename();
            m.providerSegmentName = p.parent_path().parent_path().filename();
            m.entityName = p.parent_path().filename();
            m.timestamp = IceUtil::Time::microSeconds(std::stol(p.filename()));
            return m;
        }

        void writeDataToFile(const std::filesystem::path& path, const std::vector<unsigned char>& data)
        {
            std::ofstream dataofs;
            dataofs.open(path);
            dataofs.write(reinterpret_cast<const char*>(data.data()), data.size());
            dataofs.close();
        }
    }

    EntitySnapshot::EntitySnapshot(const std::filesystem::path& p) :
        EntitySnapshotBase(getMemoryIDFromPath(p)),
        DiskStorage(p)
    {
        extractor = std::make_shared<extractor::ImageExtractor>();
        dictConverter = std::make_shared<converter::dict::JsonConverter>();
        imgConverter = std::make_shared<converter::image::PngConverter>();
    }

    std::string EntitySnapshot::getExpectedFolderName() const
    {
        return name();
    }

    void EntitySnapshot::_loadAll(armem::wm::EntitySnapshot& e) const
    {
        std::filesystem::path p = path;
        util::ensureFolderExists(p, false);

        e.id() = id();

        for (unsigned int i = 0; i < 1000; ++i) // 1000 is max size for instances in a single timestamp
        {
            if (!util::checkIfFolderExists(p / std::to_string(i)))
            {
                break;
            }

            // add instance. Do not set data, since we only return references
            e.addInstance();
        }
    }

    void EntitySnapshot::_load(armem::wm::EntitySnapshot& e) const
    {
        // Get data from disk
        std::filesystem::path p = path;
        util::ensureFolderExists(p, false);

        for (unsigned int i = 0; i < e.size(); ++i)
        {
            util::ensureFolderExists(p / std::to_string(i), false);

            std::filesystem::path data = p / std::to_string(i) / (constantes::DATA_FILENAME + dictConverter->suffix);
            std::filesystem::path metadata = p / std::to_string(i) / (constantes::METADATA_FILENAME + dictConverter->suffix);

            util::ensureFileExists(data);

            auto& ins = e.getInstance(i);

            std::ifstream dataifs(data);
            std::vector<unsigned char> datafilecontent((std::istreambuf_iterator<char>(dataifs)), (std::istreambuf_iterator<char>()));
            auto dataaron = dictConverter->convert(datafilecontent);

            std::ifstream metadataifs(metadata);
            std::vector<unsigned char> metadatafilecontent((std::istreambuf_iterator<char>(metadataifs)), (std::istreambuf_iterator<char>()));
            auto metadataaron = dictConverter->convert(metadatafilecontent);

            from_aron(metadataaron, dataaron, ins);
        }
    }

    void EntitySnapshot::_store(const armem::wm::EntitySnapshot& e) const
    {
        std::filesystem::path p = path;
        util::ensureFolderExists(p);

        for (unsigned int i = 0; i < e.size(); ++i)
        {
            std::filesystem::path instancePath = p / std::to_string(i);
            util::ensureFolderExists(instancePath);

            std::filesystem::path dataPath = instancePath / (constantes::DATA_FILENAME + dictConverter->suffix);
            std::filesystem::path metadataPath = instancePath / (constantes::METADATA_FILENAME + dictConverter->suffix);

            if (util::checkIfFileExists(dataPath) or util::checkIfFileExists(metadataPath))
            {
                continue;
            }

            auto& ins = e.getInstance(i);

            // data
            auto dataAron = std::make_shared<aron::data::Dict>();
            auto metadataAron = std::make_shared<aron::data::Dict>();
            to_aron(metadataAron, dataAron, ins);

            // extract
            auto dataExt = extractor->extract(dataAron);
            auto metadataExt = extractor->extract(metadataAron);

            // convert images
            for (const auto& [memberName, img] : dataExt.images)
            {
                std::filesystem::path imgPath = instancePath / (memberName + imgConverter->suffix);
                auto imgVec = imgConverter->convert(img);
                writeDataToFile(imgPath, imgVec);
            }

            // convert dict and metadata
            auto dataVec = dictConverter->convert(dataExt.data);
            auto metadataVec = dictConverter->convert(metadataExt.data);
            writeDataToFile(dataPath, dataVec);
            writeDataToFile(metadataPath, metadataVec);
        }
    }
}
