#include "Plugin.h"

#include <RobotAPI/libraries/armem/core/error.h>
#include <RobotAPI/libraries/armem/client/plugins/Plugin.h>

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>


namespace armarx::armem::server::plugins
{

    Plugin::~Plugin() = default;


    Plugin::Plugin(ManagedIceObject& parent, std::string prefix) :
        armarx::ComponentPlugin(parent, prefix)
    {
        addPlugin(clientPlugin);
        ARMARX_CHECK_NOT_NULL(clientPlugin);
        addPluginDependency(clientPlugin);
    }


    void Plugin::postCreatePropertyDefinitions(PropertyDefinitionsPtr& properties)
    {
        const std::string prefix = "mem.";
        if (not workingMemory.name().empty()
            and not properties->hasDefinition(prefix + "MemoryName"))
        {
            properties->optional(workingMemory.name(), prefix + "MemoryName", "Name of this memory server.");
        }

        if (longtermMemoryEnabled  // if not explicitly set to false in constructor of component
                and not properties->hasDefinition(prefix + "ltm.enabled"))
        {
            // stuff for mongodb memory
            properties->optional(longtermMemoryEnabled, prefix + "ltm.enabled");
            /*properties->optional(longtermMemory.dbsettings.host, prefix + "ltm.dbhost", "Only set if you want to have a custom host. Takes the host from 'armarx memory start' if not set.");
            properties->optional(longtermMemory.dbsettings.port, prefix + "ltm.dbport", "Only set if you want to have a custom port. Takes the port from 'armarx memory start' if not set.");
            properties->optional(longtermMemory.dbsettings.user, prefix + "ltm.dbuser", "Only set if you want to have a custom user. Takes the user from 'armarx memory start' if not set.");
            properties->optional(longtermMemory.dbsettings.password, prefix + "ltm.dbpassword", "Only set if you want to have a custom password. Takes the password from 'armarx memory start' if not set.");
            properties->optional(longtermMemory.dbsettings.database, prefix + "ltm.dbdatabase", "The name of the database you want to use for this memory. You can have multiple memories in one database.");*/

            // stuff if disk memory
            properties->optional(longtermMemory.path, "ltm.storagepath", "The path to the memory storage.");
        }

        // Publish memory updates topic
        properties->topic(memoryTopic, memoryTopicDefaultName);
    }


    void Plugin::postOnConnectComponent()
    {
        Component& parent = this->parent<Component>();

        if (clientPlugin->isMemoryNameSystemEnabled() and clientPlugin->getMemoryNameSystemClient())
        {
            registerServer(parent);
        }
        iceAdapter.setMemoryListener(memoryTopic);
    }


    void Plugin::preOnDisconnectComponent()
    {
        if (clientPlugin->isMemoryNameSystemEnabled() and clientPlugin->getMemoryNameSystemClient())
        {
            removeServer();
        }
    }


    void Plugin::setMemoryName(const std::string& memoryName)
    {
        workingMemory.name() = memoryName;
        longtermMemory.setMemoryID(workingMemory.id());
    }


    mns::dto::RegisterServerResult Plugin::registerServer(armarx::Component& parent)
    {
        MemoryID id = MemoryID().withMemoryName(workingMemory.name());
        mns::dto::MemoryServerInterfaces server;
        server.reading = ReadingMemoryInterfacePrx::uncheckedCast(parent.getProxy());
        server.writing = WritingMemoryInterfacePrx::uncheckedCast(parent.getProxy());

        mns::dto::RegisterServerResult result;
        try
        {
            clientPlugin->getMemoryNameSystemClient().registerServer(id, server);
            result.success = true;
            ARMARX_DEBUG << "Registered memory server for " << id << " in the Memory Name System (MNS).";
        }
        catch (const armem::error::ServerRegistrationOrRemovalFailed& e)
        {
            result.success = false;
            result.errorMessage = e.what();
            ARMARX_WARNING << e.what();
        }
        return result;
    }


    mns::dto::RemoveServerResult Plugin::removeServer()
    {
        MemoryID id = MemoryID().withMemoryName(workingMemory.name());

        mns::dto::RemoveServerResult result;
        try
        {
            clientPlugin->getMemoryNameSystemClient().removeServer(id);
            result.success = true;
            ARMARX_DEBUG << "Removed memory server for " << id << " from the Memory Name System (MNS).";
        }
        catch (const armem::error::ServerRegistrationOrRemovalFailed& e)
        {
            result.success = false;
            result.errorMessage = e.what();
            ARMARX_WARNING << e.what();
        }
        catch (const Ice::NotRegisteredException&)
        {
            // It's ok, the MNS is gone.
            result.success = false;
            result.errorMessage = "Memory Name System is gone.";
        }
        return result;
    }

}
