/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


// Simox
#include <SimoxUtility/algorithm/string.h>

// Header
#include "Writer.h"

// ArmarX

namespace armarx::aron::codegenerator::cpp
{
    Writer::Writer(const std::string& producerName, const std::vector<std::string>& additionalIncludesFromXMLFile) :
        CodeWriter(producerName, additionalIncludesFromXMLFile)
    {
        addSpecificWriterMethods();
        addSpecificReaderMethods();
    }

    void Writer::addSpecificWriterMethods()
    {
        // The toAron Serializer is visible by default
        {
            codegenerator::WriterInfo toAron;
            toAron.methodName = "toAron";
            toAron.returnType = "armarx::aron::data::DictPtr";
            toAron.writerClassType = "armarx::aron::data::writer::VariantWriter";
            toAron.include = "<RobotAPI/libraries/aron/core/data/rw/writer/variant/VariantWriter.h>";
            toAron.enforceConversion = "armarx::aron::data::Dict::DynamicCastAndCheck";
            dataWriters.push_back(toAron);
        }

        {
            // The toAron Serializer is visible by default
            codegenerator::WriterInfo toAronType;
            toAronType.methodName = "toAronType";
            toAronType.returnType = "armarx::aron::type::ObjectPtr";
            toAronType.writerClassType = "armarx::aron::type::writer::VariantWriter";
            toAronType.include = "<RobotAPI/libraries/aron/core/type/rw/writer/variant/VariantWriter.h>";
            toAronType.enforceConversion = "armarx::aron::type::Object::DynamicCastAndCheck";
            initialTypeWriters.push_back(toAronType);
        }

        // toJSON Method
        /*
        {
            codegenerator::WriterInfo toJSON;
            toJSON.methodName = "toJSON";
            toJSON.returnType = "nlohmann::json";
            toJSON.writerClassType = "armarx::aron::data::writer::NlohmannJSONWriter";
            toJSON.include = "<RobotAPI/libraries/aron/core/data/rw/writer/nlohmannJSON/NlohmannJSONWriter.h>";
            dataWriters.push_back(toJSON);
        }
        */
    }

    void Writer::addSpecificReaderMethods()
    {
        // The static FromAron Deserializer
        {
            codegenerator::StaticReaderInfo fromAron;
            fromAron.methodName = "FromAron";
            fromAron.argumentType = "armarx::aron::data::DictPtr";
            fromAron.returnType = OWN_TYPE_NAME;
            staticDataReaders.push_back(fromAron);
        }

        // The fromAron Deserializer is visible by default
        {
            codegenerator::ReaderInfo fromAron;
            fromAron.methodName = "fromAron";
            fromAron.argumentType = "armarx::aron::data::DictPtr";
            fromAron.readerClassType = "armarx::aron::data::reader::VariantReader";
            fromAron.include = "<RobotAPI/libraries/aron/core/data/rw/reader/variant/VariantReader.h>";
            dataReaders.push_back(fromAron);
        }

        // fromJSON Method
        /*
        {
            codegenerator::ReaderInfo fromJSON;
            fromJSON.methodName = "fromJSON";
            fromJSON.argumentType = "nlohmann::json";
            fromJSON.readerClassType = "armarx::aron::data::reader::NlohmannJSONReader";
            fromJSON.include = "<RobotAPI/libraries/aron/core/data/rw/reader/nlohmannJSON/NlohmannJSONReader.h>";
            dataReaders.push_back(fromJSON);
        }
        */
    }


    void Writer::generateTypeObjects(const std::vector<typereader::GenerateObjectInfo>& generateObjects)
    {
        for (const auto& publicGenerateObjectType : generateObjects)
        {
            const auto type = publicGenerateObjectType.correspondingType;

            //std::cout << "Generating: " << nav->getName() << std::endl;

            // Convert to Object type and create class object
            if (type == nullptr)
            {
                throw error::AronException(__PRETTY_FUNCTION__, "An received public type is null. Abort due to error!");
            }

            generator::ObjectClass generator(*type);

            CppClassPtr c = setupBasicCppClass(publicGenerateObjectType, generator);
            setupMemberFields(c, publicGenerateObjectType.doc_members, generator);

            if (type->getExtends() != nullptr)
            {
                generator::ObjectClass extendsSerializer(*type->getExtends());
                c->addInherit("public " + extendsSerializer.getFullCppTypename());
            }
            else
            {
                c->addInherit("public armarx::aron::codegenerator::cpp::AronCppClass");
                c->addInclude("<RobotAPI/libraries/aron/core/codegenerator/codewriter/cpp/AronCppClass.h>");
            }

            for (const std::string& s : additionalIncludes)
            {
                if (!s.empty())
                {
                    c->addInclude(s);
                }
            }

            // Writermethods
            for (const codegenerator::WriterInfo& info : dataWriters)
            {
                if (!info.include.empty())
                {
                    c->addInclude(info.include);
                }
                CppMethodPtr convert = generator.toSpecializedDataWriterMethod(info);
                c->addMethod(convert);
            }

            // Add methods to set the member variables
            for (const codegenerator::StaticReaderInfo& info : staticDataReaders)
            {
                CppMethodPtr convert = generator.toSpecializedStaticDataReaderMethod(info);
                c->addMethod(convert);
            }

            // Add methods to set the member variables
            for (const codegenerator::ReaderInfo& info : dataReaders)
            {
                if (!info.include.empty())
                {
                    c->addInclude(info.include);
                }
                CppMethodPtr convert = generator.toSpecializedDataReaderMethod(info);
                c->addMethod(convert);
            }

            // Typewritermethods
            for (const codegenerator::WriterInfo& info : initialTypeWriters)
            {
                if (!info.include.empty())
                {
                    c->addInclude(info.include);
                }
                CppMethodPtr convert = generator.toSpecializedTypeWriterMethod(info);
                c->addMethod(convert);
            }

            typeClasses.push_back(c);
        }
    }

    void Writer::generateTypeIntEnums(const std::vector<typereader::GenerateIntEnumInfo>& generateIntEnums)
    {
        for (const auto& publicGenerateIntEnumType : generateIntEnums)
        {
            const auto& nav = publicGenerateIntEnumType.correspondingType;
            if (nav == nullptr)
            {
                throw error::AronException(__PRETTY_FUNCTION__, "An received public type is null. Abort due to error!");
            }

            generator::IntEnumClass generator(*nav);

            CppClassPtr c = setupBasicCppClass(publicGenerateIntEnumType, generator);
            setupMemberFields(c, publicGenerateIntEnumType.doc_values, generator);

            c->addInherit("public armarx::aron::codegenerator::cpp::AronCppClass");
            c->addInclude("<RobotAPI/libraries/aron/core/codegenerator/codewriter/cpp/AronCppClass.h>");

            // ctor
            c->addCtor(generator.toCopyCtor(c->getName()));
            c->addCtor(generator.toInnerEnumCtor(c->getName()));

            // Specific methods
            CppEnumPtr enumrepresentation = generator.toInnerEnumDefinition();
            c->addInnerEnum(enumrepresentation);

            CppMethodPtr toString = generator.toToStringMethod();
            c->addMethod(toString);

            CppMethodPtr fromString = generator.toFromStringMethod();
            c->addMethod(fromString);

            CppMethodPtr intConversion = generator.toIntMethod();
            c->addMethod(intConversion);

            CppMethodPtr enumAssignment = generator.toEnumAssignmentMethod();
            c->addMethod(enumAssignment);

            CppMethodPtr enumAssignment2 = generator.toCopyAssignmentMethod();
            c->addMethod(enumAssignment2);

            CppMethodPtr enumAssignment3 = generator.toIntAssignmentMethod();
            c->addMethod(enumAssignment3);

            typeClasses.push_back(c);
        }
    }

    CppClassPtr Writer::setupBasicCppClass(const typereader::GenerateInfo& info, const Generator& gen) const
    {
        const std::string cppTypename = gen.getFullCppTypename();
        std::vector<std::string> split = simox::alg::split(cppTypename, "::");
        if (split.size() < 1)
        {
            throw error::AronException(__PRETTY_FUNCTION__, "The cpp name of an inner type was empty. Please check the type definition file if all object names are set correctly.");
        }

        std::vector<std::string> namespaces(split);
        namespaces.pop_back();

        CppClassPtr c = std::make_shared<CppClass>(namespaces, split[split.size() - 1]);

        // Add includes and guard
        std::string classDoc = "******************************************\n"
                               "* AUTOGENERATED CLASS. Please do NOT edit.\n"
                               "******************************************\n";

        if (!info.doc_author.empty() or !info.doc_brief.empty())
        {
            classDoc += (info.doc_brief.empty() ? "" : " * @brief " + info.doc_brief + "\n");
            classDoc += (info.doc_author.empty() ? "" : " * @author " + info.doc_author + "\n");
        }
        c->addClassDoc(classDoc);
        c->setPragmaOnceIncludeGuard(true);
        c->addInclude("<memory>");
        c->addInclude("<string>");
        c->addInclude("<vector>");
        c->addInclude("<map>");
        c->addInclude("<RobotAPI/interface/aron.h>");

        c->addPublicField(std::string("using ") + OWN_TYPE_NAME + " = " + cppTypename + ";");

        // ctor
        c->addCtor(gen.toCtor(c->getName()));

        // Generic methods
        //std::cout << "Generate equals method" << std::endl;
        CppMethodPtr equals = gen.toEqualsMethod();
        c->addMethod(equals);

        //std::cout << "Generate reset method" << std::endl;
        CppMethodPtr resetHard = gen.toResetHardMethod();
        c->addMethod(resetHard);

        //std::cout << "Generate init method" << std::endl;
        CppMethodPtr resetSoft = gen.toResetSoftMethod();
        c->addMethod(resetSoft);

        //std::cout << "Generate writeInit method" << std::endl;
        CppMethodPtr writeType = gen.toWriteTypeMethod();
        c->addMethod(writeType);

        //std::cout << "Generate write method" << std::endl;
        CppMethodPtr write = gen.toWriteMethod();
        c->addMethod(write);

        //std::cout << "Generate read method" << std::endl;
        CppMethodPtr read = gen.toReadMethod();
        c->addMethod(read);
        return c;
    }

    void Writer::setupMemberFields(CppClassPtr& c, const std::map<std::string, std::string>& doc_members, const generator::ObjectClass& o) const
    {
        auto publicFields = o.getPublicVariableDeclarations("");
        for (const auto& f : publicFields)
        {
            if (auto it = doc_members.find(f->getName()); it != doc_members.end())
            {
                f->setDoc("@brief " + it->second);
            }
            c->addPublicField(f);
        }
    }

    void Writer::setupMemberFields(CppClassPtr& c, const std::map<std::string, std::string>& doc_members, const generator::IntEnumClass& o) const
    {
        auto publicFields = o.getPublicVariableDeclarations("");
        for (const auto& f : publicFields)
        {
            if (auto it = doc_members.find(f->getName()); it != doc_members.end())
            {
                f->setDoc("@brief " + it->second);
            }
            c->addPublicField(f);
        }
    }
}





