/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <type_traits>

#include "../../Descriptor.h"
#include "Visitor.h"

namespace armarx::aron::data
{
    /**
     * @see type/visitor/RecursiveVisitor.h
     */
    template <class RecursiveVisitorImplementation>
    void visitRecursive(RecursiveVisitorImplementation& v, typename RecursiveVisitorImplementation::Input& o, const std::string& key, typename RecursiveVisitorImplementation::Input& po)
    {
        data::Descriptor descriptor = v.getDescriptor(o);
        switch (descriptor)
        {
            case data::Descriptor::eList:
            {
                v.visitListOnEnter(key, po, o);
                unsigned int i = 0;
                for (auto& value : v.getListElements(o))
                {
                    visitRecursive(v, value, std::to_string(i), o);
                    i++;
                }
                v.visitListOnExit(key, po, o);
                return;
            }
            case data::Descriptor::eDict:
            {
                v.visitDictOnEnter(key, po, o);
                for (auto& [key, value] : v.getDictElements(o))
                {
                    visitRecursive(v, value, key, o);
                }
                v.visitDictOnExit(key, po, o);
                return;
            }
            case data::Descriptor::eNDArray:
                return v.visitNDArray(key, po, o);
            case data::Descriptor::eInt:
                return v.visitInt(key, po, o);
            case data::Descriptor::eLong:
                return v.visitLong(key, po, o);
            case data::Descriptor::eFloat:
                return v.visitFloat(key, po, o);
            case data::Descriptor::eDouble:
                return v.visitDouble(key, po, o);
            case data::Descriptor::eString:
                return v.visitString(key, po, o);
            case data::Descriptor::eBool:
                return v.visitBool(key, po, o);
            case data::Descriptor::eUnknown:
                return v.visitUnknown(key, po, o);
        }
    }

    /**
     * @see type/visitor/RecursiveVisitor.h
     * @see data/visitor/Visitor.h
     */
    template <class RecursiveVisitorImplementation>
    void visitRecursive(RecursiveVisitorImplementation& v, typename RecursiveVisitorImplementation::DataInput& o, typename RecursiveVisitorImplementation::TypeInput& t, const std::string& key, typename RecursiveVisitorImplementation::DataInput& po, typename RecursiveVisitorImplementation::TypeInput& pt)
    {
        type::Descriptor descriptor = v.getDescriptor(t);
        switch (descriptor)
        {
            case type::Descriptor::eList:
            {
                v.visitListOnEnter(key, po, pt, o, t);
                unsigned int i = 0;
                for (auto& [value, acceptedType] : v.getListElements(o, t))
                {
                    visitRecursive(v, value, acceptedType, std::to_string(i), o, t);
                    i++;
                }
                v.visitListOnExit(key, po, pt, o, t);
                return;
            }
            case type::Descriptor::ePair:
            {
                v.visitPairOnEnter(key, po, pt, o, t);
                auto pair = v.getPairElements(o, t);
                auto first = pair.first;
                auto second = pair.second;
                visitRecursive(v, first.first, first.second, "1", o, t);
                visitRecursive(v, second.first, second.second, "2", o, t);
                v.visitPairOnExit(key, po, pt, o, t);
                return;
            }
            case type::Descriptor::eTuple:
            {
                v.visitTupleOnEnter(key, po, pt, o, t);
                unsigned int i = 0;
                for (auto& [value, acceptedType] : v.getTupleElements(o, t))
                {
                    visitRecursive(v, value, acceptedType, std::to_string(i), o, t);
                    i++;
                }
                v.visitTupleOnExit(key, po, pt, o, t);
                return;
            }
            case type::Descriptor::eDict:
            {
                    v.visitDictOnEnter(key, po, pt, o, t);
                    for (auto& [key, value, acceptedType] : v.getDictElements(o, t))
                    {
                        visitRecursive(v, value, acceptedType, key, o, t);
                    }
                    v.visitDictOnExit(key, po, pt, o, t);
                    return;
            }
            case type::Descriptor::eObject:
            {
                v.visitObjectOnEnter(key, po, pt, o, t);
                for (auto& [key, value, acceptedType] : v.getObjectElements(o, t))
                {
                    visitRecursive(v, value, acceptedType, key, o, t);
                }
                v.visitObjectOnExit(key, po, pt, o, t);
                return;
            }
            case type::Descriptor::eNDArray:
                return v.visitNDArray(key, po, pt, o, t);
            case type::Descriptor::eMatrix:
                return v.visitMatrix(key, po, pt, o, t);
            case type::Descriptor::eOrientation:
                return v.visitOrientation(key, po, pt, o, t);
            case type::Descriptor::eImage:
                return v.visitImage(key, po, pt, o, t);
            case type::Descriptor::ePointCloud:
                return v.visitPointCloud(key, po, pt, o, t);
            case type::Descriptor::ePosition:
                return v.visitPosition(key, po, pt, o, t);
            case type::Descriptor::ePose:
                return v.visitPose(key, po, pt, o, t);
            case type::Descriptor::eQuaternion:
                return v.visitQuaternion(key, po, pt, o, t);
            case type::Descriptor::eInt:
                return v.visitInt(key, po, pt, o, t);
            case type::Descriptor::eLong:
                return v.visitLong(key, po, pt, o, t);
            case type::Descriptor::eFloat:
                return v.visitFloat(key, po, pt, o, t);
            case type::Descriptor::eDouble:
                return v.visitDouble(key, po, pt, o, t);
            case type::Descriptor::eString:
                return v.visitString(key, po, pt, o, t);
            case type::Descriptor::eBool:
                return v.visitBool(key, po, pt, o, t);
            case type::Descriptor::eTime:
                return v.visitTime(key, po, pt, o, t);
            case type::Descriptor::eIntEnum:
                return v.visitIntEnum(key, po, pt, o, t);
            case type::Descriptor::eUnknown:
                return v.visitUnknown(key, po, pt, o, t);
        }
    }

    /**
     * @see type/visitor/RecursiveVisitor.h
     */
    template <class T>
    struct RecursiveVisitor : virtual public VisitorBase<T>
    {
        using Input = typename VisitorBase<T>::Input;
        using InputNonConst = typename std::remove_const<Input>::type;

        virtual std::map<std::string, InputNonConst> getDictElements(Input&) = 0;
        virtual std::vector<InputNonConst> getListElements(Input&) = 0;

        virtual void visitDictOnEnter(const std::string& key, Input& parent, Input& element) {};
        virtual void visitDictOnExit(const std::string& key, Input& parent, Input& element) {};
        virtual void visitListOnEnter(const std::string& key, Input& parent, Input& element) {};
        virtual void visitListOnExit(const std::string& key, Input& parent, Input& element) {};

        virtual void visitNDArray(const std::string& key, Input& parent, Input& element) {};
        virtual void visitInt(const std::string& key, Input& parent, Input& element) {};
        virtual void visitLong(const std::string& key, Input& parent, Input& element) {};
        virtual void visitFloat(const std::string& key, Input& parent, Input& element) {};
        virtual void visitDouble(const std::string& key, Input& parent, Input& element) {};
        virtual void visitBool(const std::string& key, Input& parent, Input& element) {};
        virtual void visitString(const std::string& key, Input& parent, Input& element) {};
        virtual void visitUnknown(const std::string& key, Input& parent, Input& element) { throw error::AronException(__PRETTY_FUNCTION__, "Unknown type in visitor."); }
        virtual ~RecursiveVisitor() = default;
    };

    /**
     * @see type/visitor/RecursiveVisitor.h
     */
    template <class T1, class T2>
    struct RecursiveTypedVisitor : virtual public TypedVisitorBase<T1, T2>
    {
        using DataInput = typename TypedVisitorBase<T1, T2>::DataInput;
        using TypeInput = typename TypedVisitorBase<T1, T2>::TypeInput;

        using TypeInputNonConst = typename std::remove_const<TypeInput>::type;
        using DataInputNonConst = typename std::remove_const<DataInput>::type;

        virtual std::map<std::string, std::pair<DataInputNonConst, TypeInputNonConst>> getObjectElements(DataInput&, TypeInput&) = 0;
        virtual std::map<std::string, std::pair<DataInputNonConst, TypeInputNonConst>> getDictElements(DataInput&, TypeInput&) = 0;
        virtual std::vector<std::pair<DataInputNonConst, TypeInputNonConst>> getListElements(DataInput&, TypeInput&) = 0;
        virtual std::pair<std::pair<DataInputNonConst, TypeInputNonConst>, std::pair<DataInputNonConst, TypeInputNonConst>> getPairElements(DataInput&, TypeInput&) = 0;
        virtual std::vector<std::pair<DataInputNonConst, TypeInputNonConst>> getTupleElements(DataInput&, TypeInput&) = 0;

        virtual void visitObjectOnEnter(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitObjectOnExit(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitDictOnEnter(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitDictOnExit(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitPairOnEnter(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitPairOnExit(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitTupleOnEnter(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitTupleOnExit(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitListOnEnter(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitListOnExit(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};

        virtual void visitMatrix(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitNDArray(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitQuaternion(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitOrientation(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitPosition(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitPose(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitImage(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitPointCloud(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitIntEnum(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitInt(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitLong(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitFloat(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitDouble(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitBool(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitString(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitTime(const std::string& key, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) {};
        virtual void visitUnknown(const std::string&, DataInput& parentData, TypeInput& parentType, DataInput& elementData, TypeInput& elementType) { throw error::AronException(__PRETTY_FUNCTION__, "Unknown type in visitor."); }
        virtual ~RecursiveTypedVisitor() = default;
    };
}
