﻿/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#pragma once

// STD/STL
#include <memory>
#include <string>

// ArmarX
#include <RobotAPI/interface/aron.h>
#include <RobotAPI/libraries/aron/core/Exception.h>


namespace armarx::aron::type
{
    /**
     * @brief The ReaderInterface class. It defines the interface to extract information from an aron representation
     */
    template <class InputType>
    class ReaderInterface
    {
    public:
        /// Extract information from an Object type
        virtual void readObject(const InputType& input, std::string& name, std::map<std::string, InputType>& memberTypes, type::Maybe& maybe) = 0;

        /// Extract information from a list type
        virtual void readList(const InputType& input, InputType& acceptedType, type::Maybe& maybe) = 0;

        /// Extract information from a dict type
        virtual void readDict(const InputType& input, InputType& acceptedType, type::Maybe& maybe) = 0;

        /// Extract information from a tuple type
        virtual void readTuple(const InputType& input, std::vector<InputType>& acceptedTypes, type::Maybe& maybe) = 0;

        /// Extract information from a pair type
        virtual void readPair(const InputType& input, std::pair<InputType, InputType>& acceptedTypes, type::Maybe& maybe) = 0;

        /// Extract information from a ndarray type
        virtual void readNDArray(const InputType& input, int& ndim, type::ndarray::ElementType& type, type::Maybe& maybe) = 0;

        /// Extract information from a matrix type
        virtual void readMatrix(const InputType& input, int& rows, int& cols, type::matrix::ElementType& type, type::Maybe& maybe) = 0;

        /// Extract information from a quaternion type
        virtual void readQuaternion(const InputType& input, type::quaternion::ElementType& type, type::Maybe& maybe) = 0;

        /// Extract information from a pointcloud type
        virtual void readPointCloud(const InputType& input, type::pointcloud::VoxelType& type, type::Maybe& maybe) = 0;

        /// Extract information from an image type
        virtual void readImage(const InputType& input, type::image::PixelType& type, type::Maybe& maybe) = 0;

        /// Extract information from a position type
        virtual void readPosition(const InputType& input, type::Maybe& maybe) = 0;

        /// Extract information from an orientation type
        virtual void readOrientation(const InputType& input, type::Maybe& maybe) = 0;

        /// Extract information from a pose type
        virtual void readPose(const InputType& input, type::Maybe& maybe) = 0;

        /// Extract information from an int enum type
        virtual void readIntEnum(const InputType& input, std::string& name, std::map<std::string, int>& acceptedValues, type::Maybe& maybe) = 0;

        /// Extract information from an int type
        virtual void readInt(const InputType& input, type::Maybe& maybe) = 0;

        /// Extract information from an long type
        virtual void readLong(const InputType& input, type::Maybe& maybe) = 0;

        /// Extract information from an float type
        virtual void readFloat(const InputType& input, type::Maybe& maybe) = 0;

        /// Extract information from an double type
        virtual void readDouble(const InputType& input, type::Maybe& maybe) = 0;

        /// Extract information from an string type
        virtual void readString(const InputType& input, type::Maybe& maybe) = 0;

        /// Extract information from an bool type
        virtual void readBool(const InputType& input, type::Maybe& maybe) = 0;

        /// Extract information from an time type
        virtual void readTime(const InputType& input, type::Maybe& maybe) = 0;
    };
}
