/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#pragma once

// STD/STL
#include <memory>
#include <string>

// ArmarX
#include <RobotAPI/interface/aron.h>
#include <RobotAPI/libraries/aron/core/Exception.h>


namespace armarx::aron::type
{
    /**
     * @brief The WriterInterface class. It defines the interface to construct aron representations (e.g. variant or nlohmann::json)
     */
    template <class ReturnType>
    class WriterInterface
    {
    public:
        /// Construct an object from the params
        virtual ReturnType writeObject(const std::string& name, const type::Maybe maybe, const std::map<std::string, ReturnType>& memberTypes, const std::optional<ReturnType>& extends) = 0;

        /// Construct a list from the params
        virtual ReturnType writeList(const type::Maybe maybe, const ReturnType& acceptedType) = 0;

        /// Construct a dict from the params
        virtual ReturnType writeDict(const type::Maybe maybe, const ReturnType& acceptedType) = 0;

        /// Construct a pair from the params
        virtual ReturnType writePair(const type::Maybe maybe, const ReturnType& acceptedType1, const ReturnType& acceptedType2) = 0;

        /// Construct a tuple from the params
        virtual ReturnType writeTuple(const type::Maybe maybe, const std::vector<ReturnType>& acceptedTypes) = 0;

        /// Construct a ndarray from the params
        virtual ReturnType writeNDArray(const int ndim, const type::ndarray::ElementType, const type::Maybe maybe) = 0;

        /// Construct a matrix from the params
        virtual ReturnType writeMatrix(const int rows, const int cols, const type::matrix::ElementType type, const type::Maybe maybe) = 0;

        /// Construct a quaternion from the params
        virtual ReturnType writeQuaternion(const type::quaternion::ElementType, const type::Maybe maybe) = 0;

        /// Construct a image from the params
        virtual ReturnType writeImage(const type::image::PixelType, const type::Maybe maybe) = 0;

        /// Construct a pointcloud from the params
        virtual ReturnType writePointCloud(const type::pointcloud::VoxelType, const type::Maybe maybe) = 0;

        /// Construct a position from the params
        virtual ReturnType writePosition(const type::Maybe maybe) = 0;

        /// Construct a orientation from the params
        virtual ReturnType writeOrientation(const type::Maybe maybe) = 0;

        /// Construct a pose from the params
        virtual ReturnType writePose(const type::Maybe maybe) = 0;

        /// Construct a int enum from the params
        virtual ReturnType writeIntEnum(const std::string& name, const std::map<std::string, int>& acceptedValues, const type::Maybe maybe) = 0;

        /// Construct a int from the params
        virtual ReturnType writeInt(const type::Maybe maybe) = 0;

        /// Construct a long from the params
        virtual ReturnType writeLong(const type::Maybe maybe) = 0;

        /// Construct a float from the params
        virtual ReturnType writeFloat(const type::Maybe maybe) = 0;

        /// Construct a double from the params
        virtual ReturnType writeDouble(const type::Maybe maybe) = 0;

        /// Construct a string from the params
        virtual ReturnType writeString(const type::Maybe maybe) = 0;

        /// Construct a bool from the params
        virtual ReturnType writeBool(const type::Maybe maybe) = 0;

        /// Construct a time from the params
        virtual ReturnType writeTime(const type::Maybe maybe) = 0;
    };
}
