/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "VariantReader.h"

// ArmarX
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/type/variant/All.h>


namespace armarx::aron::type::reader
{
    void NavigatorReader::readObject(const aron::type::VariantPtr& input, std::string& name, std::map<std::string, aron::type::VariantPtr>& memberTypes, type::Maybe& maybe)
    {
        auto o = type::Object::DynamicCastAndCheck(input);

        name = o->getObjectName();
        maybe = o->getMaybe();
        memberTypes = o->getMemberTypes();
    }
    void NavigatorReader::readList(const aron::type::VariantPtr& input, aron::type::VariantPtr& acceptedType, type::Maybe& maybe)
    {
        auto o = type::List::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
        acceptedType = o->getAcceptedType();
    }

    void NavigatorReader::readDict(const aron::type::VariantPtr& input, aron::type::VariantPtr& acceptedType, type::Maybe& maybe)
    {
        auto o = type::Dict::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
        acceptedType = o->getAcceptedType();
    }

    void NavigatorReader::readTuple(const aron::type::VariantPtr& input, std::vector<aron::type::VariantPtr>& acceptedTypes, type::Maybe& maybe)
    {
        auto o = type::Tuple::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
        acceptedTypes = o->getAcceptedTypes();
    }

    void NavigatorReader::readPair(const aron::type::VariantPtr& input, std::pair<aron::type::VariantPtr, aron::type::VariantPtr>& acceptedTypes, type::Maybe& maybe)
    {
        auto o = type::Pair::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
        acceptedTypes = o->getAcceptedTypes();
    }

    void NavigatorReader::readNDArray(const aron::type::VariantPtr& input, int& ndim, type::ndarray::ElementType& type, type::Maybe& maybe)
    {
        auto o = type::NDArray::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
        ndim = o->getNumberDimensions();
        type = o->getElementType();
    }

    void NavigatorReader::readMatrix(const aron::type::VariantPtr& input, int& rows, int& cols, type::matrix::ElementType& type, type::Maybe& maybe)
    {
        auto o = type::Matrix::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
        rows = o->getRows();
        cols = o->getCols();
        type = o->getElementType();
    }

    void NavigatorReader::readQuaternion(const aron::type::VariantPtr& input, type::quaternion::ElementType& type, type::Maybe& maybe)
    {
        auto o = type::Quaternion::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
        type = o->getElementType();
    }

    void NavigatorReader::readPointCloud(const aron::type::VariantPtr& input, type::pointcloud::VoxelType& type, type::Maybe& maybe)
    {
        auto o = type::PointCloud::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
        type = o->getVoxelType();
    }

    void NavigatorReader::readImage(const aron::type::VariantPtr& input, type::image::PixelType& type, type::Maybe& maybe)
    {
        auto o = type::Image::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
        type = o->getPixelType();
    }

    void NavigatorReader::readPosition(const aron::type::VariantPtr& input, type::Maybe& maybe)
    {
        auto o = type::Position::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
    }

    void NavigatorReader::readOrientation(const aron::type::VariantPtr& input, type::Maybe& maybe)
    {
        auto o = type::Orientation::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
    }

    void NavigatorReader::readPose(const aron::type::VariantPtr& input, type::Maybe& maybe)
    {
        auto o = type::Pose::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
    }

    void NavigatorReader::readIntEnum(const aron::type::VariantPtr& input, std::string& name, std::map<std::string, int>& acceptedValues, type::Maybe& maybe)
    {
        auto o = type::IntEnum::DynamicCastAndCheck(input);

        name = o->getEnumName();
        acceptedValues = o->getAcceptedValueMap();
        maybe = o->getMaybe();
    }

    void NavigatorReader::readInt(const aron::type::VariantPtr& input, type::Maybe& maybe)
    {
        auto o = type::Int::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
    }

    void NavigatorReader::readLong(const aron::type::VariantPtr& input, type::Maybe& maybe)
    {
        auto o = type::Long::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
    }

    void NavigatorReader::readFloat(const aron::type::VariantPtr& input, type::Maybe& maybe)
    {
        auto o = type::Float::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
    }

    void NavigatorReader::readDouble(const aron::type::VariantPtr& input, type::Maybe& maybe)
    {
        auto o = type::Double::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
    }

    void NavigatorReader::readString(const aron::type::VariantPtr& input, type::Maybe& maybe)
    {
        auto o = type::String::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
    }

    void NavigatorReader::readBool(const aron::type::VariantPtr& input, type::Maybe& maybe)
    {
        auto o = type::Bool::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
    }

    void NavigatorReader::readTime(const aron::type::VariantPtr& input, type::Maybe& maybe)
    {
        auto o = type::Time::DynamicCastAndCheck(input);

        maybe = o->getMaybe();
    }
}
