/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <type_traits>

#include "../../Descriptor.h"
#include "Visitor.h"

namespace armarx::aron::type
{
    /**
     * @brief The visitRecursive function. Calls visitX of a RecursiveVisitorImplementation recursively. For more information please see Visitor.h
     */
    template <class RecursiveVisitorImplementation>
    void visitRecursive(RecursiveVisitorImplementation& v, const std::string& key, typename RecursiveVisitorImplementation::Input& t)
    {
        auto descriptor = v.getDescriptor(t);
        switch (descriptor)
        {
            case type::Descriptor::eList:
            {
                v.visitListOnEnter(key, t);
                auto acceptedType = v.getListAcceptedType(t);
                visitRecursive(v, "acceptedType", acceptedType);
                v.visitListOnExit(key, t);
                return;
            }
            case type::Descriptor::ePair:
            {
                v.visitPairOnEnter(key, t);
                auto acceptedTypes = v.getPairAcceptedTypes(t);
                visitRecursive(v, "acceptedType1", acceptedTypes.first);
                visitRecursive(v, "acceptedType2", acceptedTypes.second);
                v.visitPairOnExit(key, t);
                return;
            }
            case type::Descriptor::eTuple:
            {
                v.visitTupleOnEnter(key, t);
                unsigned int i = 0;
                for (const auto& acceptedType : v.getTupleAcceptedTypes(t))
                {
                    visitRecursive(v, "acceptedType"+std::to_string(i), acceptedType);
                    i++;
                }
                v.visitTupleOnExit(key, t);
                return;
            }
            case type::Descriptor::eDict:
            {
                v.visitDictOnEnter(key, t);
                auto acceptedType = v.getDictAcceptedType(t);
                visitRecursive(v, "acceptedType", acceptedType);
                v.visitDictOnExit(key, t);
                return;
            }
            case type::Descriptor::eObject:
            {
                v.visitObjectOnEnter(key, t);
                for (const auto& [key, acceptedType] : v.getObjectElements(t))
                {
                    visitRecursive(v, key, acceptedType);
                }
                v.visitObjectOnExit(key, t);
                return;
            }
            case type::Descriptor::eNDArray:
                return v.visitNDArray(key, t);
            case type::Descriptor::eMatrix:
                return v.visitMatrix(key, t);
            case type::Descriptor::eOrientation:
                return v.visitOrientation(key, t);
            case type::Descriptor::eImage:
                return v.visitImage(key, t);
            case type::Descriptor::ePointCloud:
                return v.visitPointCloud(key, t);
            case type::Descriptor::ePosition:
                return v.visitPosition(key, t);
            case type::Descriptor::ePose:
                return v.visitPose(key, t);
            case type::Descriptor::eQuaternion:
                return v.visitQuaternion(key, t);
            case type::Descriptor::eInt:
                return v.visitInt(key, t);
            case type::Descriptor::eLong:
                return v.visitLong(key, t);
            case type::Descriptor::eFloat:
                return v.visitFloat(key, t);
            case type::Descriptor::eDouble:
                return v.visitDouble(key, t);
            case type::Descriptor::eString:
                return v.visitString(key, t);
            case type::Descriptor::eBool:
                return v.visitBool(key, t);
            case type::Descriptor::eTime:
                return v.visitTime(key, t);
            case type::Descriptor::eIntEnum:
                return v.visitIntEnum(key, t);
            case type::Descriptor::eUnknown:
                return v.visitUnknown(key, t);
        }
    }

    /**
     * @brief The RecursiveVisitor struct. It differs from the Visitor struct (@see Visitor.h), because it provides special visitXOnEnter and visitXOnExit methods for container types.
     * Further, it defines abstract methods to get the children from the input representation which is used by the visitRecursive method.
     */
    template <class T>
    struct RecursiveVisitor : virtual public VisitorBase<T>
    {
        using Input = typename VisitorBase<T>::Input;
        using InputNonConst = typename std::remove_const<Input>::type;

        virtual std::map<std::string, InputNonConst> getObjectAcceptedTypes(Input&) = 0;
        virtual InputNonConst getDictAcceptedType(Input&) = 0;
        virtual InputNonConst getListAcceptedType(Input&) = 0;
        virtual std::pair<InputNonConst, InputNonConst> getPairAcceptedTypes(Input&) = 0;
        virtual std::vector<InputNonConst> getTupleAcceptedTypes(Input&) = 0;

        virtual void visitObjectOnEnter(const std::string& key, Input&) {};
        virtual void visitObjectOnExit(const std::string& key, Input&) {};
        virtual void visitDictOnEnter(const std::string& key, Input&) {};
        virtual void visitDictOnExit(const std::string& key, Input&) {};
        virtual void visitPairOnEnter(const std::string& key, Input&) {};
        virtual void visitPairOnExit(const std::string& key, Input&) {};
        virtual void visitTupleOnEnter(const std::string& key, Input&) {};
        virtual void visitTupleOnExit(const std::string& key, Input&) {};
        virtual void visitListOnEnter(const std::string& key, Input&) {};
        virtual void visitListOnExit(const std::string& key, Input&) {};

        virtual void visitMatrix(const std::string& key, Input&) {};
        virtual void visitNDArray(const std::string& key, Input&) {};
        virtual void visitQuaternion(const std::string& key, Input&) {};
        virtual void visitOrientation(const std::string& key, Input&) {};
        virtual void visitPosition(const std::string& key, Input&) {};
        virtual void visitPose(const std::string& key, Input&) {};
        virtual void visitImage(const std::string& key, Input&) {};
        virtual void visitPointCloud(const std::string& key, Input&) {};
        virtual void visitIntEnum(const std::string& key, Input&) {};
        virtual void visitInt(const std::string& key, Input&) {};
        virtual void visitLong(const std::string& key, Input&) {};
        virtual void visitFloat(const std::string& key, Input&) {};
        virtual void visitDouble(const std::string& key, Input&) {};
        virtual void visitBool(const std::string& key, Input&) {};
        virtual void visitString(const std::string& key, Input&) {};
        virtual void visitTime(const std::string& key, Input&) {};
        virtual void visitUnknown(const std::string&, Input&) { throw error::AronException(__PRETTY_FUNCTION__, "Unknown type in visitor."); }
        virtual ~RecursiveVisitor() = default;
    };
}
