/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <map>
#include <string>
#include <vector>

#include "../../Descriptor.h"
#include "../../Exception.h"

namespace armarx::aron::type
{
    /**
     * @brief The visit function. Takes a visitor implementation as input and a type representation (e.g. aron::type::variant or nlohmann::json).
     * Inspired by std::visit for std::variant
     * This method only performs a single visit. This is useful if you only want to call a specific method depending on the type, or if you want to call a customized recursive function
     * If you want to call a visitX recursively, please refer to "RecursiveVisitor.h"
     */
    template <class VisitorImplementation>
    void visit(VisitorImplementation& v, typename VisitorImplementation::Input& t)
    {
        auto descriptor = v.getDescriptor(t);
        switch (descriptor)
        {
        case type::Descriptor::eObject:
            return v.visitObject(t);
        case type::Descriptor::eList:
            return v.visitList(t);
        case type::Descriptor::eDict:
            return v.visitDict(t);
        case type::Descriptor::ePair:
            return v.visitPair(t);
        case type::Descriptor::eTuple:
            return v.visitTuple(t);
        case type::Descriptor::eNDArray:
            return v.visitNDArray(t);
        case type::Descriptor::eMatrix:
            return v.visitMatrix(t);
        case type::Descriptor::eOrientation:
            return v.visitOrientation(t);
        case type::Descriptor::eImage:
            return v.visitImage(t);
        case type::Descriptor::ePointCloud:
            return v.visitPointCloud(t);
        case type::Descriptor::ePosition:
            return v.visitPosition(t);
        case type::Descriptor::ePose:
            return v.visitPose(t);
        case type::Descriptor::eQuaternion:
            return v.visitQuaternion(t);
        case type::Descriptor::eInt:
            return v.visitInt(t);
        case type::Descriptor::eLong:
            return v.visitLong(t);
        case type::Descriptor::eFloat:
            return v.visitFloat(t);
        case type::Descriptor::eDouble:
            return v.visitDouble(t);
        case type::Descriptor::eString:
            return v.visitString(t);
        case type::Descriptor::eBool:
            return v.visitBool(t);
        case type::Descriptor::eTime:
            return v.visitTime(t);
        case type::Descriptor::eIntEnum:
            return v.visitIntEnum(t);
        case type::Descriptor::eUnknown:
            return v.visitUnknown(t);
        }
    }

    /**
     * @brief The VisitorBase struct. Defines basic methods and typedefs for all visitors
     */
    template <class T>
    struct VisitorBase
    {
        using Input = T;

        virtual type::Descriptor getDescriptor(Input&) = 0;
        virtual ~VisitorBase() = default;
    };

    /**
     * @brief The Visitor struct. Defines the visitX methods a visitor has to impelement
     */
    template <class T>
    struct Visitor : virtual public VisitorBase<T>
    {
        using Input = typename VisitorBase<T>::Input;
        virtual void visitObject(Input&) {};
        virtual void visitDict(Input&) {};
        virtual void visitPair(Input&) {};
        virtual void visitTuple(Input&) {};
        virtual void visitList(Input&) {};
        virtual void visitMatrix(Input&) {};
        virtual void visitNDArray(Input&) {};
        virtual void visitQuaternion(Input&) {};
        virtual void visitOrientation(Input&) {};
        virtual void visitPosition(Input&) {};
        virtual void visitPose(Input&) {};
        virtual void visitImage(Input&) {};
        virtual void visitPointCloud(Input&) {};
        virtual void visitIntEnum(Input&) {};
        virtual void visitInt(Input&) {};
        virtual void visitLong(Input&) {};
        virtual void visitFloat(Input&) {};
        virtual void visitDouble(Input&) {};
        virtual void visitBool(Input&) {};
        virtual void visitString(Input&) {};
        virtual void visitTime(Input&) {};
        virtual void visitUnknown(Input&) { throw error::AronException(__PRETTY_FUNCTION__, "Unknown type in visitor."); }
        virtual ~Visitor() = default;
    };
}
